import io
import pydash
from aiohttp import web
from PIL import Image
from rocon_client_sdk_py.logger.rocon_logger import rocon_logger


class WebServer():
    def __init__(self, worker_client, port):
        rocon_logger.debug('start web server, port {}'.format(port))
        self._owner_worker_client = worker_client

        self._server = None
        self._runner = None

        self._port = port
        self._host_url = 'localhost'
        self._handler = WebServerRouterHandler(self._owner_worker_client)
        self._now_running = False

        self._routers = [

            web.get('/{uuid}/hello', self._handler.handle_test_hello),

            web.post('/{uuid}/controls/tick', self._handler.handle_start_worker),
            web.delete('/{uuid}/controls/tick', self._handler.handle_stop_worker),
            web.delete('/{uuid}/blackboard', self._handler.handle_reset_worker),
            web.put('/{uuid}/controls/tick', self._handler.handle_set_interval),
            web.get('/{uuid}', self._handler.handle_get_virtual_worker),
            web.get('/{uuid}/options', self._handler.handle_get_options),
            web.get('/{uuid}/blackboard', self._handler.handle_get_blackboard),
            web.get('/{uuid}/location', self._handler.handle_get_location),
            web.get('/{uuid}/controls/messages', self._handler.handle_get_messages),
            web.post('/{uuid}/controls/messages', self._handler.handle_add_messages),

            web.get('/{uuid}/register/wpm', self._handler.handle_register_at_workerpoolmanager),
            # For development mode
            web.get('/dev/{map_name}/ShowPathFinding', self._handler.handle_dev_show_path_finding),
            web.get('/dev/{map_name}/ClearPathFinding', self._handler.handle_dev_clear_path_finding),
        ]

    async def start(self):
        app = web.Application()
        app.add_routes(self._routers)

        try:
            self._runner = web.AppRunner(app)
            await self._runner.setup()

            site = web.TCPSite(self._runner, 'localhost', self._port)

            rocon_logger.debug("======= Serving on http://{}:{}/ ======".format(self._host_url, self._port))
            await site.start()
            self._now_running = True
        except Exception as err:
            rocon_logger.error('Exception occurred', exception=err)
            self._now_running = False

    @property
    def now_running(self):
        return self._now_running

class WebServerRouterHandler():
    def __init__(self, worker_client):
        self._owner_worker_client = worker_client

    async def handle_intro(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))
        return web.Response(text='Hi there!')

    async def handle_test_hello(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))
        txt = 'Hello, world'
        return web.Response(text=txt)

    def _check_valid_request(self, request):
        id = request.match_info['uuid']
        if self._owner_worker_client.uuid != id:
            rocon_logger.warning('Not matched uuid (requested uuid = {})'.format(id))
            return False

        return True

    async def handle_get_options(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))
        if self._check_valid_request(request) is False:
            return web.Response(text='Not matched uuid')

        response_data = self._owner_worker_client.worker.options
        return web.json_response(response_data)

    async def handle_get_blackboard(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))
        txt = 'handle_get_blackboard'
        return web.Response(text=txt)

    async def handle_reset_worker(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))
        if self._check_valid_request(request) is False:
            return web.Response(text='Not matched uuid')


        #TODO reset blackborad
        self._owner_worker_client.reset()

        return web.Response(text='success')

    #TODO JS 구현과 확인 필요 (all)
    async def handle_get_location(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))
        txt = 'handle_get_location'
        return web.Response(text=txt)

    async def handle_get_messages(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))
        txt = 'handle_get_messages'
        return web.Response(text=txt)

    async def handle_add_messages(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))

        if self._check_valid_request(request) is False:
            return web.Response(text='Not matched uuid')

        data = await request.json()
        name = data['name']
        body = data['body']
        overwrite = data['overwrite']

        messages = self._owner_worker_client.context.blackboard.get('recv_messages') or []
        if overwrite:
            exist_msg_idx = pydash.find_last_index(messages, {'name': name, 'receiver': self._owner_worker_client.uuid})
            if exist_msg_idx != -1:
                messages[exist_msg_idx]['body'] = body
                return web.json_response({'messages_length': len(messages)})

        new_messages = {'name': name, 'body': body, 'receiver': self._owner_worker_client.uuid}
        messages.append(new_messages)
        self._owner_worker_client.context.blackboard.set('recv_messages', messages)
        return web.json_response({'messages_length': len(messages)})


    async def handle_start_worker(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))

        if self._check_valid_request(request) is False:
            return web.Response(text='Not matched uuid')

        data = await request.json()
        interval = data['interval']

        self._owner_worker_client.start(interval)
        return web.Response(text='success')

    async def handle_stop_worker(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))

        if self._check_valid_request(request) is False:
            return web.Response(text='Not matched uuid')

        data = await request.json()
        offline = data['offline']
        #TODO set offline

        self._owner_worker_client.stop()
        return web.Response(text='success')

    async def handle_set_interval(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))

        if self._check_valid_request(request) is False:
            return web.Response(text='Not matched uuid')

        data = await request.json()
        interval = data['interval']

        self._owner_worker_client.start_tick_timer(interval)
        return web.Response(text='success')

    #TODO 정리
    async def handle_get_virtual_worker(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))

        if self._check_valid_request(request) is False:
            return web.Response(text='Not matched uuid')

        response_data = self._owner_worker_client.worker.to_json()
        return web.json_response(response_data)

    async def handle_register_at_workerpoolmanager(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))

        if self._check_valid_request(request) is False:
            return web.Response(text='Not matched uuid')

        await self._owner_worker_client.register_at_wpm_with_await()

        return web.Response(text='success')


    async def handle_dev_show_path_finding(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))
        map_name = request.match_info['map_name']
        image = self._owner_worker_client.context.map_manager.get_map_path_image_by_name(map_name)
        if image is None:

            map_names = self._owner_worker_client.context.map_manager.get_map_names()
            msg = 'Not found image, required map name : {}\nValid map names are {}'.format(map_name, map_names)
            rocon_logger.debug(msg)
            return web.Response(text=msg)
        else:
            data = self.image_to_byte_array(image)
            return web.Response(body=data, content_type='image/png')

    async def handle_dev_clear_path_finding(self, request):
        rocon_logger.debug('request url > {}'.format(request.url))
        map_name = request.match_info['map_name']

        self._owner_worker_client.context.map_manager.clear_map_path_image_by_name(map_name)

        image = self._owner_worker_client.context.map_manager.get_map_path_image_by_name(map_name)
        data = self.image_to_byte_array(image)

        return web.Response(body=data, content_type='image/png')


    def image_to_byte_array(self, image:Image):
      imgByteArr = io.BytesIO()
      image.save(imgByteArr, format=image.format)
      imgByteArr = imgByteArr.getvalue()
      return imgByteArr