import asyncio
import threading
import paho.mqtt.client as mqtt
from rocon_client_sdk_py.server.web_server import WebServer
from rocon_client_sdk_py.worker import Worker
from rocon_client_sdk_py.logger.rocon_logger import rocon_logger
from rocon_client_sdk_py.const import *
from rocon_client_sdk_py.info import info

class WorkerClient():
    def __init__(self):
        self._rocon_logger = rocon_logger
        self._rocon_logger.debug('\n', show_caller_info=False)
        self._rocon_logger.debug('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=', show_caller_info=False)
        self._rocon_logger.debug('       Start Rocon Client (Hello!)       ', show_caller_info=False)
        self._rocon_logger.debug('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=', show_caller_info=False)
        self._rocon_logger.debug('\n', show_caller_info=False)
        self._rocon_logger.debug('initializing Rocon Client')

        self._init_event_loop()

        self._mqtt_client = None
        self._worker = self.on_create(self._event_loop)
        self._worker.set_after_tick_callback(self._callback_after_tick)

        # start web server
        self._server_port = self._worker._client_configs['configs']['server']['port']
        self._web_server = WebServer(self, self._server_port)

        self._server_enabled = self._worker._client_configs['configs']['server']['enabled']
        if self._server_enabled is True:
            future = asyncio.run_coroutine_threadsafe(self._web_server.start(), self._event_loop)
            result = future.result()

            # register into WorkerPool
            self.register_at_wpm()


        self._rocon_logger.debug('initailized Rocon Client')
        self._worker.start_tick_timer(DEFAULT_TICK_INTERVAL_MS)


    @property
    def rocon_logger(self):
        return self._rocon_logger


    def _init_event_loop(self):
        self._event_loop = asyncio.get_event_loop()

        async def start():
            while True:
                #TODO check this routine
                await asyncio.sleep(1)

        def run_it_forever(loop):
            self.rocon_logger.debug('run event loop forever')
            loop.run_forever()
            self._rocon_logger.debug('stopped event loop')

        self._event_loop.create_task(start())
        self._event_thread = threading.Thread(target=run_it_forever, args=(self._event_loop,))
        self._event_thread.start()


    def _init_event_loop_old(self):
        self._event_loop = asyncio.get_event_loop()

        def f(loop):
            asyncio.set_event_loop(loop)
            loop.run_forever()
            self._rocon_logger.debug('stop event loop')

        self._event_thread = threading.Thread(target=f, args=(self._event_loop,))
        self._event_thread.start()


    @property
    def worker(self):
        """Worker 객체
        :return: 생성된 Worker 객체 리턴
        """
        return self._worker

    @property
    def uuid(self):
        """Worker의 고유한 uuid
        :return: uuid 값 리턴
        """
        return self._worker.uuid

    @property
    def context(self):
        return self._worker.context

    def run(self):

        cli_info = "[Rocon Client CLI] Enter 'q' to quilt, 'h' to help\n>"
        input_cmd = ''
        while input_cmd.lower() != 'q':
            input_cmd = input(cli_info)

            cmd = input_cmd.lower()
            if cmd == 'h':
                self._rocon_logger.cli_info("\n", show_caller_info=False)
                self._rocon_logger.cli_info("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=", show_caller_info=False)
                self._rocon_logger.cli_info("               Rocon Client CLI(Command line interface) Help                     ", show_caller_info=False)
                self._rocon_logger.cli_info("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=", show_caller_info=False)
                self._rocon_logger.cli_info("  'q' : Quit                                     ", show_caller_info=False)
                self._rocon_logger.cli_info("  'r' : Start tick-loop                          ", show_caller_info=False)
                self._rocon_logger.cli_info("  's' : Stop tick-loop                           ", show_caller_info=False)
                self._rocon_logger.cli_info("  't' : Start just one tick                      ", show_caller_info=False)
                self._rocon_logger.cli_info("  'i' : Show information                         ", show_caller_info=False)
                self._rocon_logger.cli_info("                                                 ", show_caller_info=False)
                self._rocon_logger.cli_info("  'sb' : Show log of behavior-tree's logic       ", show_caller_info=False)
                self._rocon_logger.cli_info("  'hb' : Hide log of behavior-tree's logic       ", show_caller_info=False)
                self._rocon_logger.cli_info("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=", show_caller_info=False)
                self._rocon_logger.cli_info("\n", show_caller_info=False)

            elif cmd == 't':
                self._worker.tick()
            elif cmd == 'r':
                self._worker.start_tick_timer(DEFAULT_TICK_INTERVAL_MS)
            elif cmd == 's':
                self._worker.stop_tick_timer()
            elif cmd == 'i':
                self._show_info()
            elif cmd == 'sb':
                self._rocon_logger.enable_logging_by_moduel_keyword(BT_KEYWORD)
            elif cmd == 'hb':
                self._rocon_logger.disable_logging_by_moduel_keyword(BT_KEYWORD)



        self.on_destroy()

    def on_create(self) -> Worker:
        self._rocon_logger.debug()

        return None

    def on_destroy(self):
        self._rocon_logger.debug('destroying Rocon Client')

        self.stop()

        self._event_loop.stop()
        self._event_thread.join(30)

        self._rocon_logger.debug('event loop is closed? : ' + str(self._event_loop.is_closed()))
        self._rocon_logger.debug('event loop is running? : ' + str(self._event_loop.is_running()))
        self._rocon_logger.debug('event_thread alive? : ' + str(self._event_thread.is_alive()))

        self._rocon_logger.debug('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=', show_caller_info=False)
        self._rocon_logger.debug('     Shutdown Rocon Client (Bye!)     ', show_caller_info=False)
        self._rocon_logger.debug('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=', show_caller_info=False)

    def register_at_wpm(self):

        api_pool = self._worker.context.api_worker_pool
        #TODO need to change the name of 'Agent'
        reg_data = {
            'uuid': self._worker.uuid,
            'hostWorkerClient': 'localhost',
            'portWorkerClient': self._server_port
        }


        self._worker_pool_registered = False
        
        future = asyncio.run_coroutine_threadsafe(api_pool.register_worker(self._worker.uuid, reg_data), self._event_loop)
        result = future.result()

        if result is None:
            self._rocon_logger.error('{} registration failed in WorkerPool'.format(self._worker.uuid))
        else:
            self._worker_pool_registered = True
            self._rocon_logger.debug('{} registration succeeded in WorkerPool'.format(self._worker.uuid))

    async def register_at_wpm_with_await(self):

        api_pool = self._worker.context.api_worker_pool
        # TODO need to change the name of 'Agent'
        reg_data = {
            'uuid': self._worker.uuid,
            'hostWorkerClient': 'localhost',
            'portWorkerClient': self._server_port
        }

        result = await api_pool.register_worker(self._worker.uuid, reg_data)
        if result is None:
            self._rocon_logger.error('{} registration failed in WorkerPool'.format(self._worker.uuid))
        else:
            self._worker_pool_registered = True
            self._rocon_logger.debug('{} registration succeeded in WorkerPool'.format(self._worker.uuid))


    def mqtt_on_connect(self, client, userdata, flags, rc):
        if self._server_enabled:
            try:
                self._rocon_logger.debug('connected with result code' + str(rc))
                client.subscribe('tree/')  #
            except Exception as err:
                self._rocon_logger.error('mqtt control error : {}'.format(err))

    def mqtt_on_message(self, client, userdata, msg):
        if self._server_enabled:
            self._rocon_logger.debug(msg.topic + ' ' + str(msg.payload))


    def mqtt_publish(self, payload):
        if self._server_enabled:
            try:
                if self._mqtt_client == None:
                    self.start_mqtt()

                self._mqtt_client.publish('tree/{}'.format(self._worker.uuid), payload)
            except Exception as err:
                self._rocon_logger.error('mqtt control error : {}'.format(err))


    def start_mqtt(self):
        # mqtt subscribe
        if self._mqtt_client is None:
            self.close_mqtt()

        self._mqtt_client = mqtt.Client()
        self._mqtt_client.on_connect = self.mqtt_on_connect
        self._mqtt_client.on_message = self.mqtt_on_message

        self._mqtt_client.connect('localhost', port=10997)
        self._mqtt_client.loop_start()

    def close_mqtt(self):
        if self._mqtt_client:
            self._mqtt_client.loop_stop()
            self._mqtt_client.disconnect()
            self._mqtt_client = None

        self._rocon_logger.debug('closed mqtt')

    def restart(self):
        self.stop()
        self._worker.reset()

    def start(self, interval_msec = DEFAULT_TICK_INTERVAL_MS):
        self._worker.start(interval_msec)
        self.start_mqtt()

    def stop(self, offline = False):
        self._worker.stop(offline)
        self.close_mqtt()

    def start_tick_timer(self, interval):
        self._worker.start_tick_timer(interval)

    def _callback_after_tick(self, context):
        publishing_data = context.get_trees_stringify()
        self.mqtt_publish(publishing_data)

    def _show_info(self):
        print('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=')
        print('                         Rocon Client Information                                ')
        print('                                                                                 ')
        print('                                                                  SDK ver : {}   '.format(info.version))
        print('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=')
        print(' UUID                : {}'.format(self._worker.uuid))
        print(' Hosts')
        print('  + Scheduler        : {}'.format(self._worker.context.api_report.hostname))
        print('  + Configuration    : {}'.format(self._worker.context.api_configuration.hostname))
        print('  + SiteConfig.      : {}'.format(self._worker.context.api_site_configuration.hostname))
        print('  + WorkerPool       : {}'.format(self._worker.context.api_worker_pool.hostname))
        print('')
        print(' Server')
        print('  + port             : {}'.format(self._server_port))
        print('  + enabled          : {}'.format(self._server_enabled))
        print('  + now running      : {}'.format(self._web_server.now_running))
        print('=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=')